﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MultiWiiSimulator
{
    public struct Vector
    {
        // translation vectors
        public double X, Y, Z;

        // moments of inertia
        public double Roll, Pitch, Yaw;

        public Vector(double X=0, double Y=0, double Z=0, double Roll=0, double Pitch=0, double Yaw=0)
        {
            this.X = X;
            this.Y = Y;
            this.Z = Z;
            this.Roll = Roll;
            this.Pitch = Pitch;
            this.Yaw = Yaw;
        }

/*        public Vector(Vector copy)
        {
            this.X = copy.X;
            this.Y = copy.Y;
            this.Z = copy.Z;
            this.Roll = copy.Roll;
            this.Pitch = copy.Pitch;
            this.Yaw = copy.Yaw;
        }*/
    }

    public class Copter
    {
        public static int CyclesPerSecond = 30; // number of times the physics is calculated/second
        public static double Gravity = -9.8; // meters/second^2
        public static double BaroLatency = 1.0; // seconds of latency in bar readings

        public static double SecondsOfHistory = 10.0;  // keep in seconds of history

        public static double ThrottleToThrustRatio = 40.0;
        public static double ThrustImpedance = 1.0;
        public static double TerminalVelocity = 55.0;   // approx terminal velocity of a human below-down
        public static double MaxThrustVelocity = 25.0;  // max velocity the craft is capable of pushing

        public double MaxVelocity = 15.0;   // try to stay within this velocity (m/s)
        public int MinThrottleControl = 1100;   // minimum amount of throttle when under control
        public int MaxThrottleControl = 1900;   // maximum amount of throttle when under control


        // actual location and inertia computed by the simulator
        // these numbers are not available to copter algorithms
        public Vector Position;
        public Vector Velocity;    // meters/second
        public Vector Accelleration;    // meters/second^2

        // sensor reading
        public double Baro;                 // latest baro reading
        public Vector Acc;              // accellerometer
        //public Vector EstimatedAccelleration;
        public Vector EstimatedVelocity;    // estimated velocity by deriving baro, and integrating accelleration
        public Vector EstimatedPosition;    // estimated position when predicting baro (compensate for baro latency)

        // input
        public int Throttle;     // throttle level (1000-2000)

        // motor outputs
        public double Thrust { get { return thrust; } set { thrust = Math.Max(value, 0); } }     // motor force : also in meters/second^2

        // features
        public bool AltitudeHold { get { return altitudeHold; } set { if (value != altitudeHold) { altitudeHold = value; desiredAltitude = value ? Position.Z : -1; } } }   // altitude hold on or off
        public double AltitudeKeeping { get { return desiredAltitude; } set { desiredAltitude = value; altitudeHold = true; } }
        public PID AltitudePID { get { return altitudePID; } }

        public bool VelocityHold { get { return velocityHold; } set { if(value!=velocityHold) { velocityHold = value; desiredVelocity = 0; } } }
        public double VelocityKeeping { get { return desiredVelocity; } set { desiredVelocity = Math.Min(Math.Max(value, -MaxVelocity), MaxVelocity); velocityHold = true; } }
        public PID VelocityPID { get { return velocityPID; } }

        // stats
        public long millis;
        public long Ticks { get { return ticks; } }


        #region Private Members
        private long ticks;

        private List<Vector> history;

        private double thrust;

        private PID altitudePID;
        public bool altitudeHold = false;   // altitude hold on or off
        private double desiredAltitude = -1;

        private PID velocityPID;
        private bool velocityHold = false;
        private double desiredVelocity = 0;
        #endregion

        public Copter()
        {
            // create position history and initialize all history to position 0
            history = new List<Vector>();
            //for (int i = 0, c = (int)(SecondsOfHistory * CyclesPerSecond); i < c; i++)
            //    history.Add(new Vector());

            //Position = new Vector();
            //Velocity = new Vector();
            //EstimatedVelocity = new Vector();
            //EstimatedPosition = new Vector();
            //Accelleration = new Vector();
            //Acc = new Vector();

            altitudePID = new PID(1.0, 0.0, 0.0);
            velocityPID = new PID(1.0, 0.1, 0.0);
        }

        public void SetAltitude(double new_altitude)
        {
            desiredAltitude = new_altitude;
            altitudeHold = true;
        }

        // simulate real world variables
        public void Tick()
        {
            while (history.Count > SecondsOfHistory * CyclesPerSecond)
                history.RemoveAt(0);
            //history.Add(new Vector(Position));
            history.Add(Position);

            Vector previousVelocity = Velocity;

            // if bar enabled, compute throttle output
            if (AltitudeHold)
                ComputeAltitudeHold();
            if (VelocityHold || AltitudeHold)
                ComputeVelocityHold();

            // convert throttle input to thrust
            double motor_input = (Throttle - 1000) / ThrottleToThrustRatio;
            double ThrustError = motor_input - Thrust;
            Thrust += ThrustError / ThrustImpedance;

            // estimate some drag falling based on closeness of velocity to terminal velocity
            double drag = 0;
            if (false && Velocity.Z < 0)
            {
                drag = Gravity * (Velocity.Z / TerminalVelocity);
                if (drag > -Gravity) drag = -Gravity;
            }

            // remove gravity from velocity
            Accelleration.Z = Thrust + Gravity + drag;

            Velocity.Z += Accelleration.Z/CyclesPerSecond;
            if (Velocity.Z > MaxThrustVelocity) Velocity.Z = MaxThrustVelocity;

            // add inertia into location
            Position.Z += Velocity.Z/CyclesPerSecond;


            // check to see if we hit ground
            if (Position.Z <= 0 && (Accelleration.Z<0 || Velocity.Z<0))
            {
                Position.Z = 0;
                Velocity.Z = 0;
                Accelleration.Z = 0;
                Acc.Z = 512;
            }

            // update sensors
            SimSensors();
            ComputeEstimates();

            ticks++;
        }


        private void SimSensors()
        {
            int baro_history_index = (int)(BaroLatency * CyclesPerSecond);
            double lastbaro = Baro;
            if (history.Count > baro_history_index + 1)
                Baro = history[history.Count - baro_history_index - 1].Z;
            else
                Baro = 0;

            // simulate ACC sensor
            // we will take the current accelleration, 0 actual accelleration equals gravity (value 512)
            // 0 would equal positive 1G
            //double acc = (Velocity.Z - previousVelocity.Z) * CyclesPerSecond;
            double acc = Accelleration.Z - Gravity;
            Acc.Z = acc / Gravity * -512;
        }

        private long lbt;
        private double lbm;
        private double vnot;
        private void ComputeEstimates()
        {
            // extrapolate velocity
            // EstimatedVelocity.Z = (Baro - lastbaro) * CyclesPerSecond;  // remember this value is over BaroLatency old!
            if (true && (lbt - millis > 1000))
            {
                double ev = (Baro - lbm) * CyclesPerSecond;
                double velerror = ev - vnot;
                EstimatedVelocity.Z += velerror;
                lbm = Baro;
                lbt = millis;
                vnot = EstimatedVelocity.Z;
            }

            // add in our accell to our estimated velocity
            double realaccel = Acc.Z / -512 * Gravity + Gravity;
            EstimatedVelocity.Z += realaccel / CyclesPerSecond;

            // extrapolate estimated true position by compensating for bar latency
            EstimatedPosition.Z = history[history.Count - 1].Z;
            //EstimatedPosition.Z += EstimatedVelocity.Z * BaroLatency;
        }

        private void ComputeVelocityHold()
        {
            if(velocityPID==null || !VelocityHold) return;
            //double velerror = Velocity.Z - desiredVelocity;
            double velerror = EstimatedVelocity.Z - desiredVelocity;

            // heuristic: feed-forward the expected gravity error
            velerror += Gravity/CyclesPerSecond;

            double output = velocityPID.Compute(velerror);
            int throttle = (int)(1000 - output*100);
            throttle = Math.Max(Math.Min(throttle, MaxThrottleControl), MinThrottleControl);
            Throttle = throttle;
        }

        private void ComputeAltitudeHold()
        {
            if(!altitudeHold || desiredAltitude<0) return;
            double alterror = EstimatedPosition.Z - desiredAltitude;
            double output = -altitudePID.Compute(alterror);
            VelocityKeeping = output;
        }
    }
}
