﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.Threading;

namespace MultiWii
{
    public class MultiWiiLink
    {
        public string Port { get { return port; } set { port = value; } }
        public int BaudRate { get { return baudrate; } set { baudrate=value;} }
        public static string RequestDataString = "J";
        
        public static string EnableLogString = "L";
        public static string DisableLogString = "l";

        private const int LogMessageLength = 30;

        public bool PollMode = false;

        #region Private Members
        private SerialPort dev;
        
        private string port;
        private int baudrate = 115200;
        private int errors = 0;
        private long last_ts=0;
        private int resyncs = 0;
        #endregion


        public void Connect()
        {
            //try
            //{
                dev = new SerialPort(port, baudrate, Parity.None, 8, StopBits.One);
                dev.DtrEnable = false;
                dev.Handshake = Handshake.None;
                dev.Open();
                dev.NewLine = "\n";
                //dev.ReadTimeout = 2000;
                if(!PollMode)
                    dev.Write(EnableLogString);
            //}
            //catch (System.IO.IOException ex)
            //{
            //    return;
            //}
        }

        public void Disconnect()
        {
            if (dev.IsOpen)
            {
                dev.Write(DisableLogString);
                dev.Close();
            }
        }

        private void RecoverSerialState()
        {
            // read until end of line
            var s = dev.ReadLine();
            Console.WriteLine(resyncs++);
        }

        public SensorData Read()
        {
            // "L" ACC[XYZ] GYRO[XYZ] MAG[XYZ] ALT HEADING MILLIS_H:MILLIS_L "\n"
            if (dev == null)
                return null;

            try
            {
                if (PollMode)
                    dev.Write(RequestDataString);
                byte[] data = new byte[125];
                if (dev.Read(data, 0, LogMessageLength) == LogMessageLength)
                {
                    // confirm we have a valid sensor data message
                    if (data[0] != 'L' && data[LogMessageLength-1] != '\n')
                    {
                        RecoverSerialState();
                        return Read();
                    }

#if false
                    var sensors=SensorData.ReadBrief(data, 1);
                    if (last_ts != 0 && sensors.TimeStamp < last_ts)
                        return Read();  // out of time order, read error, read again
                    else
                    {
                        last_ts = sensors.TimeStamp;
                        return sensors;
                    }
#else
                    return SensorData.ReadBrief(data, 1);
#endif
                }
                return null;
            }
            catch (TimeoutException ex)
            {
                // no response, send another request and wait
                if (!PollMode)
                    dev.Write(EnableLogString);
                return Read();
            } 
            catch (System.IO.IOException ex)
            {
                errors++;
                if (errors < 25 && !dev.IsOpen)
                {
                    Connect();
                    return Read();
                }
                else
                    return null;
            }
            catch (InvalidOperationException ex)
            {
                errors++;
                if (errors < 25 && !dev.IsOpen)
                {
                    Connect();
                    return Read();
                }
                else
                    return null;
            }
        }

        public double[] ReadArray()
        {
            // "L" ACC[XYZ] GYRO[XYZ] MAG[XYZ] ALT HEADING MILLIS_H:MILLIS_L "\n"
            var sensors = Read();
            var values = new double[] { 0,0,0,  0,0,0,  0,0,0,  0,0,  0,0 };

            if(sensors!=null) {
                values[0] = (double)sensors.TimeStamp/1000000;
                values[1] = sensors.Acc.Roll;
                values[2] = sensors.Acc.Pitch;
                values[3] = sensors.Acc.Z;

                values[4] = sensors.Gyro.Roll;
                values[5] = sensors.Gyro.Pitch;
                values[6] = sensors.Gyro.Yaw;

                values[7] = sensors.Mag.Roll;
                values[8] = sensors.Mag.Pitch;
                values[9] = sensors.Mag.Yaw;

                values[10] = sensors.VerticalAcc;
                values[11] = sensors.Altitude;
                values[12]  = sensors.Heading;
            }
            return values;
        }

        public short[] ReadDirect(int words)
        {
            // "L" ACC[XYZ] GYRO[XYZ] MAG[XYZ] ALT HEADING MILLIS_H:MILLIS_L "\n"
            if (dev == null)
                return null;
            
            try
            {
                if (PollMode)
                    dev.Write(RequestDataString);
                byte[] data = new byte[125];
                int len, timeout=1000;
                while (dev.BytesToRead < words * 2 + 2 && --timeout>0)
                    Thread.Sleep(1);
                if ((len=dev.Read(data, 0, words*2+2)) == words*2+2)
                {
                    // confirm we have a valid sensor data message
                    if (data[0] != 'L' || data[words*2+1] != '\n')
                    {
                        RecoverSerialState();
                        return ReadDirect(words);
                    }

                    return SensorData.ReadInt16s(data, 1, words);
                }
                return null;
            }
            catch (TimeoutException ex)
            {
                // no response, send another request and wait
                if (!PollMode)
                    dev.Write(EnableLogString);
                return new short[0];
            }
            catch (System.IO.IOException ex)
            {
                errors++;
                if (errors < 25 && !dev.IsOpen)
                {
                    Connect();
                    return new short[0];
                }
                else
                    return null;
            }
            catch (InvalidOperationException ex)
            {
                errors++;
                if (errors < 25 && !dev.IsOpen)
                {
                    Connect();
                    return new short[0];
                }
                else
                    return null;
            }
        }

        public void WriteParameter8(char param_name, byte value)
        {
            if (dev == null)
                return;
            byte[] data = new byte[3];
            data[0] = (byte)param_name;
            data[1] = value;
            dev.Write(data, 0, 3);
        }

        public void WriteParameter16(char param_name, short value)
        {
            if (dev == null)
                return;
            byte[] data = new byte[3];
            data[0] = (byte)param_name;
            data[1] = (byte)(value & 0xff);
            data[2] = (byte)((value >> 8) & 0xff);
            dev.Write(data, 0, 3);
        }

        public void WriteParameter32(char param_name, int value)
        {
            if (dev == null)
                return;
            byte[] data = new byte[3];
            data[0] = (byte)param_name;
            data[1] = (byte)(value & 0xff);
            data[2] = (byte)((value >> 8) & 0xff);
            data[3] = (byte)((value >> 16) & 0xff);
            data[4] = (byte)((value >> 24) & 0xff);
            dev.Write(data, 0, 5);
        }
    }
}
